/*
 * eth_ecat_datagram.hpp
 *
 *  Created on: May 2, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_

#include <cstdint>

#include "ethernet/eth.hpp"
#include "ethernet_industry/ethercattype.hpp"
#include "ethernet_industry/eth_ecat_types.h"
#include "ethernet_industry/eth_ecat_command.hpp"

namespace free_rtos {

namespace datagram {

using TEcatWkc = uint16_t;

template<typename EcatDgDataT>
struct EcatDg {
public:
    EcatDg(uint8_t cmd, uint32_t address, EcatDgDataT& data, ec_moredatagrams more = ec_moredatagrams::EC_MOREDATAGRAMS_LAST, uint8_t idx = 0, uint16_t irq = 0)
    : header_{
        cmd,
        idx,
        address,
        {.bits = {
                  .len = sizeof(EcatDgDataT),
                  .c = static_cast<uint16_t>(ec_circframe::EC_CIRCFRAME_NOT),
                  .m = static_cast<uint16_t>(more)}},
        irq}
     , data_{data}
     , wkc_{0} { }

    EcatDg(uint8_t cmd, uint32_t address, ec_moredatagrams more = ec_moredatagrams::EC_MOREDATAGRAMS_LAST, uint8_t idx = 0, uint16_t irq = 0)
    : header_{
      cmd,
      idx,
      address,
      {.bits = {
                .len = sizeof(EcatDgDataT),
                .c = static_cast<uint16_t>(ec_circframe::EC_CIRCFRAME_NOT),
                .m = static_cast<uint16_t>(more)}},
      irq}
    , wkc_{0} { }

    EcatDg() { }

    TEcatDgHeader& get_header() {
        return header_;
    }

    EcatDgDataT& get_data() {
        return data_;
    }

    TEcatWkc get_wkc() {
        return wkc_;
    }

private:
    TEcatDgHeader header_;
    EcatDgDataT data_;
    TEcatWkc wkc_;

} __attribute__ ((packed));

class IEcatDatagram {
public:
    IEcatDatagram(ec_moredatagrams more, TEcatWkc wkc)
    : more_{more}
    , wkc_{wkc}
    { }

    IEcatDatagram() { }

    IEcatDatagram& operator+(IEcatDatagram &next) {
        more_ = ec_moredatagrams::EC_MOREDATAGRAMS_MORE;
        next_ = &next;

        return next;
    }

    IEcatDatagram* get_next() {
        return next_;
    }

    IEcatDatagram& set_next(IEcatDatagram &next) {

        return operator+(next);
    }

    virtual uint8_t* pack(uint8_t *raw) = 0;
    virtual uint8_t* unpack(uint8_t *raw) = 0;

    TEcatWkc get_wkc() {
        return wkc_;
    }

protected:
    ec_moredatagrams more_;
    TEcatWkc wkc_;

private:
    IEcatDatagram *next_{nullptr};
};

template<typename CommandT, typename EcatDgDataT>
class EcatDatagram : public IEcatDatagram {
    static_assert(std::is_base_of<command::EcatCommandBase, CommandT>::value == true, "CommandT should be derived from ECatCommandBase");

public:
    EcatDatagram(CommandT command, EcatDgDataT &data_out, EcatDgDataT &data_in)
    : IEcatDatagram{ec_moredatagrams::EC_MOREDATAGRAMS_LAST, 0x0000}
    , command_{command}
    , data_out_{&data_out}
    , data_in_{&data_in} { }

    EcatDatagram(CommandT command, EcatDgDataT &data)
    : IEcatDatagram{ec_moredatagrams::EC_MOREDATAGRAMS_LAST, 0x0000}
    , command_{command}
    , data_out_{&data}
    , data_in_{&data} { }

    EcatDatagram(CommandT command)
    : IEcatDatagram{ec_moredatagrams::EC_MOREDATAGRAMS_LAST, 0x0000}
    , command_{command} { }

    EcatDatagram() { }

    virtual uint8_t* pack(uint8_t *raw) override {
        EcatDg<EcatDgDataT> *p_dg;

        if(data_out_ != nullptr) {
           p_dg = new(raw) EcatDg<EcatDgDataT> {command_.get_cmd(), command_.get_address(), *data_out_, more_};
        }else{
           p_dg = new(raw) EcatDg<EcatDgDataT> {command_.get_cmd(), command_.get_address(), more_};
        }

        (void)p_dg;

        return (raw + sizeof(EcatDg<EcatDgDataT>));
    }

    virtual uint8_t* unpack(uint8_t *raw) override {
        EcatDg<EcatDgDataT> *p_dg  = reinterpret_cast<EcatDg<EcatDgDataT>*>(raw);

        if(data_in_ != nullptr) {
            *data_in_ = p_dg->get_data();
        }

        wkc_ = p_dg->get_wkc();

        return (raw + sizeof(EcatDg<EcatDgDataT>));
    }

private:
    CommandT command_;
    EcatDgDataT *data_out_{nullptr};
    EcatDgDataT *data_in_{nullptr};
};

class EcatTelegram : public Handler {
public:
    EcatTelegram(Eth& eth)
    : eth_{eth}
    , tx_flow_{*eth.getTxFlowPtr()} {
          eth_.getEthStackPtr()->Register(ETH_PROT_ECAT_LE, this);
    }

    virtual int32_t Process(uint8_t *p_data, uint32_t len) override;

    void init(TEthMacPorts port_id) {
        port_id_ = port_id;
    }

    void transfer(IEcatDatagram& first);

private:
    Eth& eth_;
    EthTxFlowIface& tx_flow_;
    TEthMacPorts port_id_;

    free_rtos::Semaphore rx_sem_;

    TEthPkt buffer_out_;
    TEthPkt buffer_in_;

    void pack(IEcatDatagram& first);
    void unpack(IEcatDatagram& first);
};

}


}



#endif /* FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_ */
