/*
 * eth_ecat_mailbox.hpp
 *
 *  Created on: May 3, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_MAILBOX_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_MAILBOX_HPP_

#include <vector>

#include "ethernet_industry/eth_ecat.hpp"

namespace free_rtos {

struct BufferProperties {
    uint16_t address;
    uint16_t length;
};

struct SyncManager {
    address::Offset offset;
    uint32_t default_setting;
};

struct SyncManagerProperties {
    BufferProperties mbx;
    uint32_t setting;
};

enum class SyncManagerDirection : uint8_t {
   READ = 0,
   WRITE
};

class EcatMailboxSlave {
public:
    EcatMailboxSlave(EcatSlave& slave)
    : slave_(slave) { }

    template<typename TypeT>
    void read_mailbox_info_from_eeprom(eeprom::EEPROM& eeprom) {
        auto slave_address = slave_.get_slave_address<TypeT>();

        eeprom.read<TypeT>(slave_address, ECT_SII_TXMBXADR, mbx_[static_cast<size_t>(SyncManagerDirection::READ)]);
        eeprom.read<TypeT>(slave_address, ECT_SII_RXMBXADR, mbx_[static_cast<size_t>(SyncManagerDirection::WRITE)]);

        eeprom.read<TypeT>(slave_address, ECT_PDOOUTPUTADR, pdo_[0]);
        eeprom.read<TypeT>(slave_address, ECT_PDOINPUTADR, pdo_[1]);

        DebugP_log("mbx_[READ] = 0x%04x\r\n", mbx_[0]);
        DebugP_log("mbx_[WRITE] = 0x%04x\r\n", mbx_[1]);

        DebugP_log("pdo_[OUTPUT] = 0x%04x\r\n", pdo_[0]);
        DebugP_log("pdo_[INPUT] = 0x%04x\r\n", pdo_[1]);
    }

    template<typename TypeT>
    void register_sync_manager(datagram::EcatTelegram& telegram, sync_manager sm_read, sync_manager sm_write) {
        using TCommand = command::EcatCommand<TypeT, command::WR>;

        static constexpr std::array<SyncManager, 4> syncManagers = {{
            {ECT_REG_SM0, EC_DEFAULTMBXSM0},
            {ECT_REG_SM1, EC_DEFAULTMBXSM1},
            {ECT_REG_SM2, EC_DEFAULTMBXSM2},
            {ECT_REG_SM3, EC_DEFAULTMBXSM3}
        }};

        auto slave_address = slave_.get_slave_address<TypeT>();

        BufferProperties& mbx_read = mbx_[static_cast<size_t>(SyncManagerDirection::READ)];
        SyncManager sync_manager_read = syncManagers[static_cast<size_t>(sm_read)];
        SyncManagerProperties properties_read{mbx_read, sync_manager_read.default_setting};
        datagram::EcatDatagram<TCommand, SyncManagerProperties> datagram_read{ {{slave_address, sync_manager_read.offset}}, properties_read };

        BufferProperties& mbx_write = mbx_[static_cast<size_t>(SyncManagerDirection::WRITE)];
        SyncManager sync_manager_write = syncManagers[static_cast<size_t>(sm_write)];
        SyncManagerProperties properties_write{mbx_write, sync_manager_write.default_setting};
        datagram::EcatDatagram<TCommand, SyncManagerProperties> datagram_write{ {{slave_address, sync_manager_write.offset}}, properties_write };

        datagram_read + datagram_write;

        telegram.transfer(datagram_read);

        DebugP_log("datagram_read.get_wkc() = %d\r\n", datagram_read.get_wkc());
        DebugP_log("datagram_write.get_wkc() = %d\r\n", datagram_write.get_wkc());
    }

private:
    std::array<BufferProperties, 2> mbx_;
    std::array<BufferProperties, 2> pdo_;

    EcatSlave& slave_;
};

class EthEcatMailbox {
public:
    EthEcatMailbox(EthEcat& ecat): ecat_{ecat} { }

    void mailbox_registration();
    void sync_manager_registration();

    uint16_t config_init();

private:
    EthEcat& ecat_;

    std::vector<EcatMailboxSlave> mailbox_slaves_;
};

}

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_MAILBOX_HPP_ */
